from __future__ import annotations

import importlib.util
import math
import unittest
from pathlib import Path


def load_plot_module():
    root = Path(__file__).resolve().parents[2]
    module_path = root / "src" / "ezjww" / "plot.py"
    spec = importlib.util.spec_from_file_location("ezjww_plot_module", module_path)
    if spec is None or spec.loader is None:
        raise RuntimeError(f"failed to load module from {module_path}")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


PLOT = load_plot_module()


class PlotHelperTests(unittest.TestCase):
    def test_normalize_layer_filter(self):
        self.assertIsNone(PLOT._normalize_layer_filter(None))
        self.assertEqual(PLOT._normalize_layer_filter([" A ", "", "B"]), {"A", "B"})
        self.assertIsNone(PLOT._normalize_layer_filter([" ", ""]))

    def test_line_style_mapping(self):
        self.assertEqual(PLOT._line_style("CONTINUOUS"), "-")
        self.assertEqual(PLOT._line_style("dashed"), "--")
        self.assertEqual(PLOT._line_style("DASHDOT"), "-.")
        self.assertEqual(PLOT._line_style("DOT"), ":")
        self.assertEqual(PLOT._line_style("unknown"), "-")

    def test_ellipse_points_endpoints_for_full_loop(self):
        points = PLOT._ellipse_points(10.0, 5.0, 3.0, 0.0, 0.5, 0.0, 2.0 * math.pi)
        self.assertGreaterEqual(len(points), 24)
        self.assertAlmostEqual(points[0][0], points[-1][0], places=6)
        self.assertAlmostEqual(points[0][1], points[-1][1], places=6)

    def test_aci_to_color(self):
        self.assertEqual(PLOT._aci_to_color(1), "#ff0000")
        self.assertEqual(PLOT._aci_to_color(7), "#000000")
        fallback = PLOT._aci_to_color(200)
        self.assertIsInstance(fallback, tuple)
        self.assertEqual(len(fallback), 3)


if __name__ == "__main__":
    unittest.main()



from typing import TypedDict


class LayerHeader(TypedDict):
    state: int
    protect: int
    name: str


class LayerGroupHeader(TypedDict):
    state: int
    write_layer: int
    scale: float
    protect: int
    name: str
    layers: list[LayerHeader]


class JwwHeader(TypedDict):
    version: int
    memo: str
    paper_size: int
    write_layer_group: int
    layer_groups: list[LayerGroupHeader]


class EntityBase(TypedDict):
    group: int
    pen_style: int
    pen_color: int
    pen_width: int
    layer: int
    layer_group: int
    flag: int


class LinePayload(TypedDict):
    start_x: float
    start_y: float
    end_x: float
    end_y: float


class PointPayload(TypedDict):
    x: float
    y: float
    is_temporary: bool
    code: int
    angle: float
    scale: float


class TextPayload(TypedDict):
    start_x: float
    start_y: float
    end_x: float
    end_y: float
    text_type: int
    size_x: float
    size_y: float
    spacing: float
    angle: float
    font_name: str
    content: str


class JwwEntity(TypedDict, total=False):
    type: str
    base: EntityBase
    start_x: float
    start_y: float
    end_x: float
    end_y: float
    center_x: float
    center_y: float
    radius: float
    start_angle: float
    arc_angle: float
    tilt_angle: float
    flatness: float
    is_full_circle: bool
    x: float
    y: float
    is_temporary: bool
    code: int
    angle: float
    scale: float
    text_type: int
    size_x: float
    size_y: float
    spacing: float
    font_name: str
    content: str
    point1_x: float
    point1_y: float
    point2_x: float
    point2_y: float
    point3_x: float
    point3_y: float
    point4_x: float
    point4_y: float
    color: int | None
    ref_x: float
    ref_y: float
    scale_x: float
    scale_y: float
    rotation: float
    def_number: int
    block_name: str | None
    line: LinePayload
    text: TextPayload
    sxf_mode: int | None
    aux_lines: list[LinePayload]
    aux_points: list[PointPayload]


class BlockDef(TypedDict):
    number: int
    is_referenced: bool
    name: str
    base: EntityBase
    entities: list[JwwEntity]


class BlockReferenceValidation(TypedDict):
    total_references: int
    resolved_references: int
    unresolved_def_numbers: list[int]
    has_unresolved: bool


class JwwDocument(TypedDict):
    header: JwwHeader
    entities: list[JwwEntity]
    block_defs: list[BlockDef]
    block_def_names: dict[int, str]
    entity_counts: dict[str, int]
    validation: BlockReferenceValidation


class DxfLayer(TypedDict):
    name: str
    color: int
    line_type: str
    frozen: bool
    locked: bool


class DxfEntity(TypedDict, total=False):
    type: str
    layer: str
    color: int
    line_type: str
    x1: float
    y1: float
    x2: float
    y2: float
    center_x: float
    center_y: float
    radius: float
    start_angle: float
    end_angle: float
    major_axis_x: float
    major_axis_y: float
    minor_ratio: float
    start_param: float
    end_param: float
    x: float
    y: float
    height: float
    rotation: float
    content: str
    style: str
    x3: float
    y3: float
    x4: float
    y4: float
    block_name: str
    scale_x: float
    scale_y: float


class DxfBlock(TypedDict):
    name: str
    base_x: float
    base_y: float
    entities: list[DxfEntity]


class DxfDocument(TypedDict):
    layers: list[DxfLayer]
    entities: list[DxfEntity]
    blocks: list[DxfBlock]
    unsupported_entities: list[str]


def hello_from_bin() -> str: ...
def is_jww_file(path: str) -> bool: ...
def read_header(path: str) -> JwwHeader: ...
def read_document(path: str) -> JwwDocument: ...
def read_dxf_document(
    path: str,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> DxfDocument: ...
def read_dxf_string(
    path: str,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> str: ...
def write_dxf(
    path: str,
    output_path: str,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> None: ...



from __future__ import annotations

import json
import sys
import tempfile
import unittest
from io import StringIO
from pathlib import Path
from unittest.mock import patch

ROOT = Path(__file__).resolve().parents[2]
SRC = ROOT / "src"
try:
    import ezjww
except ModuleNotFoundError:
    if str(SRC) not in sys.path:
        sys.path.insert(0, str(SRC))
    import ezjww


def sample_path() -> Path:
    return ROOT / "jww_samples" / "Test1.jww"


class PublicApiTests(unittest.TestCase):
    def test_readfile_modelspace_query(self):
        drawing = ezjww.readfile(sample_path())
        msp = drawing.modelspace()
        self.assertGreater(len(msp), 0)
        lines = msp.query("LINE")
        self.assertGreater(len(lines), 0)
        self.assertTrue(all(e.get("type") == "LINE" for e in lines))

    def test_modelspace_query_selector_multi_type(self):
        drawing = ezjww.readfile(sample_path())
        msp = drawing.modelspace()
        selected = msp.query("LINE POINT")
        self.assertGreater(len(selected), 0)
        self.assertTrue(all(e.get("type") in {"LINE", "POINT"} for e in selected))
        self.assertEqual(len(selected), len(msp.query("LINE")) + len(msp.query("POINT")))

    def test_modelspace_query_selector_filters(self):
        drawing = ezjww.readfile(sample_path())
        msp = drawing.modelspace()
        with_filters = msp.query('LINE[layer=="#lv4", color==5]')
        self.assertGreater(len(with_filters), 0)
        self.assertTrue(all(e.get("type") == "LINE" for e in with_filters))
        self.assertTrue(all(e.get("layer") == "#lv4" for e in with_filters))
        self.assertTrue(all(e.get("color") == 5 for e in with_filters))

        all_on_layer = msp.query('[layer=="#lv4"]')
        self.assertGreaterEqual(len(all_on_layer), len(with_filters))
        self.assertTrue(all(e.get("layer") == "#lv4" for e in all_on_layer))

    def test_modelspace_query_rejects_invalid_selector(self):
        drawing = ezjww.readfile(sample_path())
        msp = drawing.modelspace()
        with self.assertRaises(ValueError):
            msp.query("LINE[layer~=5]")

    def test_new_drawing_defaults(self):
        drawing = ezjww.new()
        self.assertEqual(len(drawing.modelspace()), 0)
        audit = drawing.audit()
        self.assertFalse(audit["has_issues"])
        self.assertEqual(audit["unsupported_entities"], [])
        self.assertEqual(audit["unresolved_def_numbers"], [])

    def test_audit_from_path(self):
        result = ezjww.audit(sample_path())
        self.assertIn("has_issues", result)
        self.assertIn("issue_codes", result)
        self.assertIn("unresolved_count", result)
        self.assertIn("unsupported_count", result)
        self.assertIn("unsupported_entities", result)
        self.assertIn("unresolved_def_numbers", result)
        self.assertEqual(result["unresolved_def_numbers"], [])
        self.assertEqual(result["issue_codes"], [])
        self.assertEqual(result["unresolved_count"], 0)
        self.assertEqual(result["unsupported_count"], 0)

    def test_bbox_from_path(self):
        result = ezjww.bbox(sample_path())
        self.assertIsNotNone(result)
        assert result is not None
        self.assertIn("min_x", result)
        self.assertIn("min_y", result)
        self.assertIn("max_x", result)
        self.assertIn("max_y", result)
        self.assertIn("width", result)
        self.assertIn("height", result)
        self.assertIn("entity_count", result)
        self.assertGreater(result["width"], 0.0)
        self.assertGreater(result["height"], 0.0)
        self.assertGreater(result["entity_count"], 0)

    def test_bbox_for_new_drawing_is_none(self):
        drawing = ezjww.new()
        self.assertIsNone(drawing.bbox())

    def test_stats_from_path(self):
        result = ezjww.stats(sample_path())
        self.assertIn("entity_count", result)
        self.assertIn("type_count", result)
        self.assertIn("layer_count", result)
        self.assertIn("color_count", result)
        self.assertIn("by_type", result)
        self.assertIn("by_layer", result)
        self.assertIn("by_color", result)
        self.assertGreater(result["entity_count"], 0)
        self.assertIn("LINE", result["by_type"])

    def test_modelspace_stats_for_subset(self):
        drawing = ezjww.readfile(sample_path())
        lines = drawing.modelspace().query("LINE")
        result = ezjww.Modelspace(lines).stats()
        self.assertEqual(set(result["by_type"].keys()), {"LINE"})
        self.assertEqual(result["entity_count"], len(lines))

    def test_report_from_path(self):
        result = ezjww.report(sample_path())
        self.assertIn("source_path", result)
        self.assertIn("explode_inserts", result)
        self.assertIn("max_block_nesting", result)
        self.assertIn("audit", result)
        self.assertIn("bbox", result)
        self.assertIn("stats", result)
        self.assertIn("has_issues", result["audit"])
        self.assertIn("entity_count", result["stats"])

    def test_to_dxf_string_from_path(self):
        text = ezjww.to_dxf_string(sample_path())
        self.assertIn("SECTION", text)
        self.assertTrue(text.endswith("  0\nEOF\n"))

    def test_drawing_to_dxf_string_with_options(self):
        drawing = ezjww.readfile(sample_path())
        text = drawing.to_dxf_string(explode_inserts=True, max_block_nesting=16)
        self.assertIn("ENTITIES", text)
        self.assertTrue(text.endswith("  0\nEOF\n"))

    def test_drawing_to_dxf_string_rejects_new_drawing(self):
        drawing = ezjww.new()
        with self.assertRaises(ValueError):
            drawing.to_dxf_string()

    def test_to_dxf_accepts_explode_options(self):
        drawing = ezjww.readfile(sample_path())
        regular = drawing.to_dxf()
        exploded = drawing.to_dxf(explode_inserts=True, max_block_nesting=16)
        self.assertIn("entities", regular)
        self.assertIn("entities", exploded)

    def test_to_dxf_rejects_invalid_max_block_nesting(self):
        drawing = ezjww.readfile(sample_path())
        with self.assertRaises(ValueError):
            drawing.to_dxf(explode_inserts=True, max_block_nesting=0)

    def test_cli_to_dxf_report_json(self):
        with tempfile.TemporaryDirectory(prefix="ezjww_test_") as tmp_dir:
            tmp = Path(tmp_dir)
            dxf_out = tmp / "out.dxf"
            report_out = tmp / "report.json"
            code = ezjww._run(
                [
                    "to-dxf",
                    str(sample_path()),
                    "-o",
                    str(dxf_out),
                    "--report",
                    "json",
                    "--report-path",
                    str(report_out),
                ]
            )
            self.assertEqual(code, 0)
            self.assertTrue(dxf_out.exists())
            self.assertTrue(report_out.exists())
            report = json.loads(report_out.read_text(encoding="utf-8"))
            self.assertTrue(report["ok"])
            self.assertIn("audit", report)
            self.assertIn("explode_inserts", report)
            self.assertIn("max_block_nesting", report)

    def test_print_json_fallback_for_non_utf8_stdout(self):
        class _AsciiStdout:
            def __init__(self):
                self.parts: list[str] = []

            def write(self, text: str) -> int:
                text.encode("ascii")
                self.parts.append(text)
                return len(text)

            def flush(self) -> None:
                return None

            def getvalue(self) -> str:
                return "".join(self.parts)

        fake_stdout = _AsciiStdout()
        with patch("sys.stdout", new=fake_stdout):
            ezjww._print_json({"text": "日本語"})

        parsed = json.loads(fake_stdout.getvalue())
        self.assertEqual(parsed["text"], "日本語")

    def test_cli_audit_json(self):
        buf = StringIO()
        with patch("sys.stdout", new=buf):
            code = ezjww._run(["audit", str(sample_path()), "--json"])
        self.assertEqual(code, 0)
        out = json.loads(buf.getvalue())
        self.assertIn("has_issues", out)
        self.assertIn("issue_codes", out)
        self.assertIn("unsupported_count", out)
        self.assertIn("unresolved_count", out)
        self.assertFalse(out["has_issues"])

    def test_cli_audit_fail_on_issues(self):
        fake = {
            "source_path": "dummy.jww",
            "total_references": 1,
            "resolved_references": 0,
            "unresolved_def_numbers": [10],
            "unresolved_count": 1,
            "unsupported_entities": [],
            "unsupported_count": 0,
            "issue_codes": ["UNRESOLVED_BLOCK_REFERENCES"],
            "has_issues": True,
            "warnings": ["unresolved block references detected"],
        }
        with patch.object(ezjww, "audit", return_value=fake):
            code = ezjww._run(["audit", str(sample_path()), "--fail-on-issues"])
        self.assertEqual(code, 3)

    def test_cli_audit_rejects_invalid_max_block_nesting(self):
        code = ezjww._run(["audit", str(sample_path()), "--max-block-nesting", "0"])
        self.assertEqual(code, 2)

    def test_cli_bbox_json(self):
        buf = StringIO()
        with patch("sys.stdout", new=buf):
            code = ezjww._run(["bbox", str(sample_path()), "--json", "--explode-inserts"])
        self.assertEqual(code, 0)
        out = json.loads(buf.getvalue())
        self.assertIn("min_x", out)
        self.assertIn("min_y", out)
        self.assertIn("max_x", out)
        self.assertIn("max_y", out)
        self.assertIn("width", out)
        self.assertIn("height", out)
        self.assertIn("entity_count", out)

    def test_cli_bbox_rejects_invalid_max_block_nesting(self):
        code = ezjww._run(["bbox", str(sample_path()), "--max-block-nesting", "0"])
        self.assertEqual(code, 2)

    def test_cli_stats_json(self):
        buf = StringIO()
        with patch("sys.stdout", new=buf):
            code = ezjww._run(["stats", str(sample_path()), "--json"])
        self.assertEqual(code, 0)
        out = json.loads(buf.getvalue())
        self.assertIn("entity_count", out)
        self.assertIn("by_type", out)
        self.assertIn("by_layer", out)
        self.assertIn("by_color", out)
        self.assertGreater(out["entity_count"], 0)

    def test_cli_stats_rejects_invalid_max_block_nesting(self):
        code = ezjww._run(["stats", str(sample_path()), "--max-block-nesting", "0"])
        self.assertEqual(code, 2)

    def test_cli_report_json(self):
        buf = StringIO()
        with patch("sys.stdout", new=buf):
            code = ezjww._run(["report", str(sample_path()), "--json"])
        self.assertEqual(code, 0)
        out = json.loads(buf.getvalue())
        self.assertIn("audit", out)
        self.assertIn("bbox", out)
        self.assertIn("stats", out)
        self.assertIn("has_issues", out["audit"])

    def test_cli_report_fail_on_issues(self):
        fake = {
            "source_path": "dummy.jww",
            "explode_inserts": False,
            "max_block_nesting": 32,
            "audit": {"has_issues": True},
            "bbox": None,
            "stats": {"entity_count": 0},
        }
        with patch.object(ezjww, "report", return_value=fake):
            code = ezjww._run(["report", str(sample_path()), "--fail-on-issues"])
        self.assertEqual(code, 3)

    def test_cli_report_rejects_invalid_max_block_nesting(self):
        code = ezjww._run(["report", str(sample_path()), "--max-block-nesting", "0"])
        self.assertEqual(code, 2)

    def test_cli_to_dxf_rejects_invalid_max_block_nesting(self):
        with tempfile.TemporaryDirectory(prefix="ezjww_test_") as tmp_dir:
            tmp = Path(tmp_dir)
            dxf_out = tmp / "out.dxf"
            code = ezjww._run(
                [
                    "to-dxf",
                    str(sample_path()),
                    "-o",
                    str(dxf_out),
                    "--max-block-nesting",
                    "0",
                ]
            )
            self.assertEqual(code, 2)
            self.assertFalse(dxf_out.exists())

    def test_cli_to_dxf_dir_report_json(self):
        with tempfile.TemporaryDirectory(prefix="ezjww_test_") as tmp_dir:
            tmp = Path(tmp_dir)
            out_dir = tmp / "dxf"
            report_out = tmp / "dir_report.json"
            code = ezjww._run(
                [
                    "to-dxf-dir",
                    str(ROOT / "jww_samples"),
                    "-o",
                    str(out_dir),
                    "--report",
                    "json",
                    "--report-path",
                    str(report_out),
                ]
            )
            self.assertEqual(code, 0)
            self.assertTrue(report_out.exists())
            report = json.loads(report_out.read_text(encoding="utf-8"))
            self.assertEqual(report["failed"], 0)
            self.assertGreater(report["converted"], 0)
            self.assertEqual(len(report["items"]), report["converted"])
            self.assertIn("explode_inserts", report)
            self.assertIn("max_block_nesting", report)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import math
from pathlib import Path
from typing import Any, Iterable


def _load_matplotlib() -> tuple[Any, Any]:
    try:
        import matplotlib.pyplot as plt
        from matplotlib import patches
    except Exception as exc:  # pragma: no cover - runtime dependency path
        raise ImportError(
            "matplotlib is required for plotting. Install with: pip install 'ezjww[plot]'"
        ) from exc
    return plt, patches


def _normalize_layer_filter(layers: Iterable[str] | None) -> set[str] | None:
    if layers is None:
        return None
    normalized = {layer.strip() for layer in layers if layer.strip()}
    return normalized or None


def _aci_to_color(aci: int) -> Any:
    mapping = {
        1: "#ff0000",
        2: "#ffff00",
        3: "#00ff00",
        4: "#00ffff",
        5: "#0000ff",
        6: "#ff00ff",
        7: "#000000",
        8: "#808080",
        9: "#c0c0c0",
    }
    if aci in mapping:
        return mapping[aci]
    if aci <= 0 or aci == 256:
        return "#000000"
    hue = (aci % 255) / 255.0
    # fallback color generation for extended ACI indexes
    return (hue, 0.7, 0.9)


def _line_style(line_type: str) -> str:
    name = line_type.upper()
    if name == "CONTINUOUS":
        return "-"
    if name in {"DASHED", "DASHED2"}:
        return "--"
    if name == "DASHDOT":
        return "-."
    if name == "DOT":
        return ":"
    return "-"


def _ellipse_points(
    center_x: float,
    center_y: float,
    major_axis_x: float,
    major_axis_y: float,
    minor_ratio: float,
    start_param: float,
    end_param: float,
) -> list[tuple[float, float]]:
    start = start_param
    end = end_param
    if end <= start:
        end += 2.0 * math.pi
    span = end - start
    count = max(24, int(64 * (span / (2.0 * math.pi))))

    u_x = major_axis_x
    u_y = major_axis_y
    v_x = -major_axis_y * minor_ratio
    v_y = major_axis_x * minor_ratio

    points = []
    for i in range(count + 1):
        t = start + (span * i / count)
        cos_t = math.cos(t)
        sin_t = math.sin(t)
        x = center_x + u_x * cos_t + v_x * sin_t
        y = center_y + u_y * cos_t + v_y * sin_t
        points.append((x, y))
    return points


def plot_dxf_document(
    dxf_document: dict[str, Any],
    *,
    ax: Any | None = None,
    layers: Iterable[str] | None = None,
    linewidth: float = 0.8,
    point_size: float = 12.0,
    draw_text: bool = True,
    draw_points: bool = True,
    draw_inserts: bool = True,
    text_scale: float = 1.0,
    invert_y: bool = False,
    equal_aspect: bool = True,
    autoscale: bool = True,
    show: bool = False,
    save_path: str | Path | None = None,
    dpi: int = 150,
    figsize: tuple[float, float] = (10.0, 10.0),
) -> Any:
    plt, patches = _load_matplotlib()
    layer_filter = _normalize_layer_filter(layers)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.figure

    for entity in dxf_document.get("entities", []):
        layer = str(entity.get("layer", "0"))
        if layer_filter is not None and layer not in layer_filter:
            continue

        entity_type = str(entity.get("type", ""))
        color = _aci_to_color(int(entity.get("color", 256)))
        line_style = _line_style(str(entity.get("line_type", "CONTINUOUS")))

        if entity_type == "LINE":
            ax.plot(
                [entity["x1"], entity["x2"]],
                [entity["y1"], entity["y2"]],
                color=color,
                linewidth=linewidth,
                linestyle=line_style,
            )
        elif entity_type == "CIRCLE":
            patch = patches.Circle(
                (entity["center_x"], entity["center_y"]),
                entity["radius"],
                fill=False,
                edgecolor=color,
                linewidth=linewidth,
                linestyle=line_style,
            )
            ax.add_patch(patch)
        elif entity_type == "ARC":
            start = float(entity["start_angle"])
            end = float(entity["end_angle"])
            if end < start:
                end += 360.0
            patch = patches.Arc(
                (entity["center_x"], entity["center_y"]),
                2.0 * entity["radius"],
                2.0 * entity["radius"],
                angle=0.0,
                theta1=start,
                theta2=end,
                edgecolor=color,
                linewidth=linewidth,
                linestyle=line_style,
            )
            ax.add_patch(patch)
        elif entity_type == "ELLIPSE":
            major_x = float(entity["major_axis_x"])
            major_y = float(entity["major_axis_y"])
            major_radius = math.hypot(major_x, major_y)
            minor_ratio = float(entity["minor_ratio"])
            start_param = float(entity["start_param"])
            end_param = float(entity["end_param"])

            if major_radius <= 0.0:
                continue

            span = end_param - start_param
            if span <= 0.0:
                span += 2.0 * math.pi

            is_full = abs(span - 2.0 * math.pi) < 1e-6
            if is_full:
                angle_deg = math.degrees(math.atan2(major_y, major_x))
                patch = patches.Ellipse(
                    (entity["center_x"], entity["center_y"]),
                    width=2.0 * major_radius,
                    height=2.0 * major_radius * minor_ratio,
                    angle=angle_deg,
                    fill=False,
                    edgecolor=color,
                    linewidth=linewidth,
                    linestyle=line_style,
                )
                ax.add_patch(patch)
            else:
                points = _ellipse_points(
                    float(entity["center_x"]),
                    float(entity["center_y"]),
                    major_x,
                    major_y,
                    minor_ratio,
                    start_param,
                    end_param,
                )
                xs, ys = zip(*points)
                ax.plot(xs, ys, color=color, linewidth=linewidth, linestyle=line_style)
        elif entity_type == "POINT":
            if draw_points:
                ax.scatter([entity["x"]], [entity["y"]], s=point_size, c=[color], marker="o")
        elif entity_type == "TEXT":
            if draw_text:
                content = str(entity.get("content", ""))
                height = max(6.0, float(entity.get("height", 2.5)) * text_scale)
                ax.text(
                    entity["x"],
                    entity["y"],
                    content,
                    color=color,
                    fontsize=height,
                    rotation=float(entity.get("rotation", 0.0)),
                    ha="left",
                    va="bottom",
                )
        elif entity_type == "SOLID":
            points = [
                (entity["x1"], entity["y1"]),
                (entity["x2"], entity["y2"]),
                (entity["x3"], entity["y3"]),
                (entity["x4"], entity["y4"]),
            ]
            patch = patches.Polygon(
                points,
                closed=True,
                facecolor=color,
                edgecolor=color,
                linewidth=max(0.3, linewidth * 0.8),
                alpha=0.3,
            )
            ax.add_patch(patch)
        elif entity_type == "INSERT":
            if draw_inserts:
                x = float(entity.get("x", 0.0))
                y = float(entity.get("y", 0.0))
                ax.scatter([x], [y], s=point_size * 1.5, c=[color], marker="x")
                name = str(entity.get("block_name", ""))
                if name:
                    ax.text(x, y, name, color=color, fontsize=7.0, ha="left", va="bottom")

    if equal_aspect:
        ax.set_aspect("equal", adjustable="datalim")
    if autoscale:
        ax.autoscale_view()
    if invert_y:
        ax.invert_yaxis()

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_title("JWW Plot")
    ax.grid(False)

    if save_path is not None:
        output = Path(save_path)
        output.parent.mkdir(parents=True, exist_ok=True)
        fig.savefig(output, dpi=dpi, bbox_inches="tight")
    if show:
        plt.show()

    return ax


def plot_jww(
    path: str | Path,
    *,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
    **kwargs: Any,
) -> Any:
    from ezjww._core import read_dxf_document

    dxf_document = read_dxf_document(
        str(path),
        explode_inserts,
        max_block_nesting,
    )
    return plot_dxf_document(dxf_document, **kwargs)



from __future__ import annotations

import argparse
import json
import math
import re
import sys
from collections import Counter
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from ezjww._core import (
    hello_from_bin,
    is_jww_file,
    read_document,
    read_dxf_document,
    read_dxf_string,
    read_header,
    write_dxf,
)
from ezjww.plot import plot_dxf_document, plot_jww

__all__ = [
    "Drawing",
    "Modelspace",
    "audit",
    "bbox",
    "hello_from_bin",
    "is_jww_file",
    "new",
    "readfile",
    "read_header",
    "read_document",
    "read_dxf_document",
    "read_dxf_string",
    "to_dxf_string",
    "write_dxf",
    "plot_dxf_document",
    "plot_jww",
    "report",
    "stats",
]


@dataclass
class Modelspace:
    entities: list[dict[str, Any]]

    def __iter__(self):
        return iter(self.entities)

    def __len__(self) -> int:
        return len(self.entities)

    def query(
        self,
        entity_type: str | None = None,
        *,
        layer: str | None = None,
        color: int | None = None,
    ) -> list[dict[str, Any]]:
        query_types, query_layer, query_color = _parse_query_selector(entity_type)
        if layer is None:
            layer = query_layer
        if color is None:
            color = query_color

        matched = self.entities
        if query_types is not None:
            matched = [e for e in matched if str(e.get("type", "")).upper() in query_types]
        if layer is not None:
            matched = [e for e in matched if str(e.get("layer", "")) == layer]
        if color is not None:
            matched = [e for e in matched if int(e.get("color", -1)) == color]
        return matched

    def bbox(self) -> dict[str, float | int] | None:
        return _dxf_bbox({"entities": self.entities})

    def stats(self) -> dict[str, Any]:
        return _dxf_stats({"entities": self.entities})


class Drawing:
    def __init__(
        self,
        *,
        source_path: str | None,
        jww_document: dict[str, Any] | None = None,
        dxf_document: dict[str, Any] | None = None,
    ) -> None:
        self._source_path = source_path
        self._jww_document = jww_document
        self._dxf_cache: dict[tuple[bool, int], dict[str, Any]] = {}
        if dxf_document is not None:
            self._dxf_cache[(False, 32)] = dxf_document

    @classmethod
    def from_file(cls, path: str | Path) -> "Drawing":
        source = str(path)
        jww_document = read_document(source)
        return cls(source_path=source, jww_document=jww_document)

    @classmethod
    def new(cls) -> "Drawing":
        return cls(
            source_path=None,
            jww_document=None,
            dxf_document={
                "layers": [],
                "entities": [],
                "blocks": [],
                "unsupported_entities": [],
            },
        )

    @property
    def source_path(self) -> str | None:
        return self._source_path

    @property
    def header(self) -> dict[str, Any] | None:
        if self._jww_document is None:
            return None
        return self._jww_document.get("header")

    @property
    def jww_document(self) -> dict[str, Any] | None:
        return self._jww_document

    def to_dxf(
        self,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> dict[str, Any]:
        nesting = _normalize_max_block_nesting(max_block_nesting)
        key = (bool(explode_inserts), nesting)
        if key not in self._dxf_cache:
            if self._source_path is None:
                self._dxf_cache[key] = {
                    "layers": [],
                    "entities": [],
                    "blocks": [],
                    "unsupported_entities": [],
                }
            else:
                self._dxf_cache[key] = read_dxf_document(
                    self._source_path,
                    explode_inserts,
                    nesting,
                )
        return self._dxf_cache[key]

    def modelspace(
        self,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> Modelspace:
        return Modelspace(
            self.to_dxf(
                explode_inserts=explode_inserts,
                max_block_nesting=max_block_nesting,
            ).get("entities", [])
        )

    def audit(
        self,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> dict[str, Any]:
        dxf = self.to_dxf(
            explode_inserts=explode_inserts,
            max_block_nesting=max_block_nesting,
        )
        unsupported = list(dxf.get("unsupported_entities", []))

        unresolved: list[int] = []
        total_refs = 0
        resolved_refs = 0
        if self._jww_document is not None:
            validation = self._jww_document.get("validation", {})
            unresolved = list(validation.get("unresolved_def_numbers", []))
            total_refs = int(validation.get("total_references", 0))
            resolved_refs = int(validation.get("resolved_references", 0))

        warnings: list[str] = []
        issue_codes: list[str] = []
        if unresolved:
            warnings.append("unresolved block references detected")
            issue_codes.append("UNRESOLVED_BLOCK_REFERENCES")
        if unsupported:
            warnings.append("unsupported entities exist for DXF conversion")
            issue_codes.append("UNSUPPORTED_DXF_ENTITIES")

        return {
            "source_path": self._source_path,
            "total_references": total_refs,
            "resolved_references": resolved_refs,
            "unresolved_def_numbers": unresolved,
            "unresolved_count": len(unresolved),
            "unsupported_entities": unsupported,
            "unsupported_count": len(unsupported),
            "issue_codes": issue_codes,
            "has_issues": bool(unresolved or unsupported),
            "warnings": warnings,
        }

    def bbox(
        self,
        *,
        explode_inserts: bool = True,
        max_block_nesting: int = 32,
    ) -> dict[str, float | int] | None:
        dxf = self.to_dxf(
            explode_inserts=explode_inserts,
            max_block_nesting=max_block_nesting,
        )
        return _dxf_bbox(dxf)

    def stats(
        self,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> dict[str, Any]:
        dxf = self.to_dxf(
            explode_inserts=explode_inserts,
            max_block_nesting=max_block_nesting,
        )
        return _dxf_stats(dxf)

    def report(
        self,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> dict[str, Any]:
        nesting = _normalize_max_block_nesting(max_block_nesting)
        dxf = self.to_dxf(
            explode_inserts=explode_inserts,
            max_block_nesting=nesting,
        )
        return {
            "source_path": self._source_path,
            "explode_inserts": bool(explode_inserts),
            "max_block_nesting": int(nesting),
            "audit": self.audit(
                explode_inserts=explode_inserts,
                max_block_nesting=nesting,
            ),
            "bbox": _dxf_bbox(dxf),
            "stats": _dxf_stats(dxf),
        }

    def to_dxf_string(
        self,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> str:
        if self._source_path is None:
            raise ValueError(
                "to_dxf_string() requires a source-backed drawing. use readfile(path)."
            )
        nesting = _normalize_max_block_nesting(max_block_nesting)
        return read_dxf_string(
            self._source_path,
            explode_inserts,
            nesting,
        )

    def saveas(
        self,
        output_path: str | Path,
        *,
        explode_inserts: bool = False,
        max_block_nesting: int = 32,
    ) -> None:
        if self._source_path is None:
            raise ValueError("saveas() requires a source-backed drawing. use readfile(path).")
        nesting = _normalize_max_block_nesting(max_block_nesting)
        write_dxf(
            self._source_path,
            str(output_path),
            explode_inserts,
            nesting,
        )

    def plot(
        self,
        *,
        explode_inserts: bool = True,
        max_block_nesting: int = 32,
        **kwargs: Any,
    ):
        nesting = _normalize_max_block_nesting(max_block_nesting)
        return plot_dxf_document(
            self.to_dxf(
                explode_inserts=explode_inserts,
                max_block_nesting=nesting,
            ),
            **kwargs,
        )


def readfile(path: str | Path) -> Drawing:
    return Drawing.from_file(path)


def new() -> Drawing:
    return Drawing.new()


def audit(
    path: str | Path,
    *,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> dict[str, Any]:
    return readfile(path).audit(
        explode_inserts=explode_inserts,
        max_block_nesting=max_block_nesting,
    )


def bbox(
    path: str | Path,
    *,
    explode_inserts: bool = True,
    max_block_nesting: int = 32,
) -> dict[str, float | int] | None:
    return readfile(path).bbox(
        explode_inserts=explode_inserts,
        max_block_nesting=max_block_nesting,
    )


def to_dxf_string(
    path: str | Path,
    *,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> str:
    nesting = _normalize_max_block_nesting(max_block_nesting)
    return read_dxf_string(
        str(path),
        explode_inserts,
        nesting,
    )


def stats(
    path: str | Path,
    *,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> dict[str, Any]:
    return readfile(path).stats(
        explode_inserts=explode_inserts,
        max_block_nesting=max_block_nesting,
    )


def report(
    path: str | Path,
    *,
    explode_inserts: bool = False,
    max_block_nesting: int = 32,
) -> dict[str, Any]:
    return readfile(path).report(
        explode_inserts=explode_inserts,
        max_block_nesting=max_block_nesting,
    )


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="ezjww")
    subparsers = parser.add_subparsers(dest="command", required=True)

    audit_cmd = subparsers.add_parser("audit", help="run conversion-oriented health checks")
    audit_cmd.add_argument("path", help="input .jww file")
    audit_cmd.add_argument("--json", action="store_true", help="print audit result as JSON")
    audit_cmd.add_argument(
        "--fail-on-issues",
        action="store_true",
        help="exit with non-zero status when issues are detected",
    )
    audit_cmd.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references before DXF-side checks",
    )
    audit_cmd.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    bbox_cmd = subparsers.add_parser("bbox", help="calculate drawing extents")
    bbox_cmd.add_argument("path", help="input .jww file")
    bbox_cmd.add_argument("--json", action="store_true", help="print extents as JSON")
    bbox_cmd.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references before extents calculation",
    )
    bbox_cmd.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    stats_cmd = subparsers.add_parser("stats", help="show entity distribution statistics")
    stats_cmd.add_argument("path", help="input .jww file")
    stats_cmd.add_argument("--json", action="store_true", help="print statistics as JSON")
    stats_cmd.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references before statistics calculation",
    )
    stats_cmd.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    report_cmd = subparsers.add_parser("report", help="emit combined audit+bbox+stats report")
    report_cmd.add_argument("path", help="input .jww file")
    report_cmd.add_argument("--json", action="store_true", help="print report as JSON")
    report_cmd.add_argument(
        "--fail-on-issues",
        action="store_true",
        help="exit with non-zero status when issues are detected",
    )
    report_cmd.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references before report generation",
    )
    report_cmd.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    info = subparsers.add_parser("info", help="show JWW summary")
    info.add_argument("path", help="input .jww file")
    info.add_argument("--json", action="store_true", help="print as JSON")

    to_dxf = subparsers.add_parser("to-dxf", help="convert single JWW to DXF")
    to_dxf.add_argument("path", help="input .jww file")
    to_dxf.add_argument("-o", "--output", help="output .dxf path (default: input stem + .dxf)")
    to_dxf.add_argument(
        "--report",
        choices=["json"],
        help="emit conversion report in the selected format",
    )
    to_dxf.add_argument(
        "--report-path",
        help="path to write conversion report (default: stdout)",
    )
    to_dxf.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references into transformed primitive entities",
    )
    to_dxf.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    to_dxf_dir = subparsers.add_parser(
        "to-dxf-dir", help="convert all .jww files in a directory"
    )
    to_dxf_dir.add_argument("input_dir", help="directory containing .jww files")
    to_dxf_dir.add_argument("-o", "--output-dir", help="output directory")
    to_dxf_dir.add_argument(
        "-r", "--recursive", action="store_true", help="scan subdirectories recursively"
    )
    to_dxf_dir.add_argument(
        "--fail-fast", action="store_true", help="stop at first conversion error"
    )
    to_dxf_dir.add_argument(
        "--report",
        choices=["json"],
        help="emit conversion report in the selected format",
    )
    to_dxf_dir.add_argument(
        "--report-path",
        help="path to write conversion report (default: stdout)",
    )
    to_dxf_dir.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references into transformed primitive entities",
    )
    to_dxf_dir.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    plot = subparsers.add_parser("plot", help="render JWW to image with matplotlib")
    plot.add_argument("path", help="input .jww file")
    plot.add_argument(
        "-o",
        "--output",
        help="output image path (default: input stem + .png, unless --show only)",
    )
    plot.add_argument("--show", action="store_true", help="show interactive plot window")
    plot.add_argument(
        "--layers",
        help="comma-separated layer names to draw (e.g. '0,1,2' or '#lv4,#lv5')",
    )
    plot.add_argument("--dpi", type=int, default=150, help="output image DPI")
    plot.add_argument("--invert-y", action="store_true", help="invert Y axis")
    plot.add_argument("--no-text", action="store_true", help="hide TEXT entities")
    plot.add_argument("--no-points", action="store_true", help="hide POINT entities")
    plot.add_argument("--no-inserts", action="store_true", help="hide INSERT markers")
    plot.add_argument(
        "--explode-inserts",
        action="store_true",
        help="expand INSERT references into transformed primitive entities before plotting",
    )
    plot.add_argument(
        "--max-block-nesting",
        type=int,
        default=32,
        help="maximum block nesting depth for INSERT expansion",
    )

    return parser


def _collect_jww_files(input_dir: Path, recursive: bool) -> list[Path]:
    if recursive:
        iterator = input_dir.rglob("*")
    else:
        iterator = input_dir.iterdir()
    return sorted(
        path
        for path in iterator
        if path.is_file() and path.suffix.lower() == ".jww"
    )


def _default_output_path(input_path: Path) -> Path:
    return input_path.with_suffix(".dxf")


_SELECTOR_TOKEN_RE = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")
_LAYER_FILTER_RE = re.compile(r"""^layer\s*==\s*(?P<quote>['"])(?P<value>.*?)(?P=quote)$""", re.IGNORECASE)
_COLOR_FILTER_RE = re.compile(r"^color\s*==\s*(?P<value>-?\d+)$", re.IGNORECASE)
_FILTER_SPLIT_RE = re.compile(r"\s*(?:,|&&|\band\b)\s*", re.IGNORECASE)


def _parse_query_selector(selector: str | None) -> tuple[set[str] | None, str | None, int | None]:
    if selector is None:
        return None, None, None
    if not isinstance(selector, str):
        raise TypeError("query selector must be a string or None")

    text = selector.strip()
    if not text:
        return None, None, None

    if "[" in text:
        if text.count("[") != 1 or not text.endswith("]"):
            raise ValueError(f"invalid query selector: {selector!r}")
        entity_part, filter_part = text.split("[", 1)
        filter_expr = filter_part[:-1].strip()
    else:
        entity_part = text
        filter_expr = ""

    query_types = _parse_selector_types(entity_part.strip(), selector)
    query_layer, query_color = _parse_selector_filters(filter_expr, selector)
    return query_types, query_layer, query_color


def _parse_selector_types(entity_part: str, original_selector: str) -> set[str] | None:
    if not entity_part or entity_part == "*":
        return None

    tokens = entity_part.replace(",", " ").split()
    if not tokens:
        return None

    types: set[str] = set()
    for token in tokens:
        upper = token.upper()
        if upper == "*":
            if len(tokens) > 1:
                raise ValueError(f"invalid query selector: {original_selector!r}")
            return None
        if not _SELECTOR_TOKEN_RE.fullmatch(token):
            raise ValueError(f"invalid query selector: {original_selector!r}")
        types.add(upper)
    return types


def _parse_selector_filters(filter_expr: str, original_selector: str) -> tuple[str | None, int | None]:
    if not filter_expr:
        return None, None

    layer: str | None = None
    color: int | None = None
    parts = [part.strip() for part in _FILTER_SPLIT_RE.split(filter_expr) if part.strip()]
    if not parts:
        raise ValueError(f"invalid query selector: {original_selector!r}")

    for part in parts:
        layer_match = _LAYER_FILTER_RE.fullmatch(part)
        if layer_match is not None:
            if layer is not None:
                raise ValueError(f"duplicate layer filter in selector: {original_selector!r}")
            layer = layer_match.group("value")
            continue

        color_match = _COLOR_FILTER_RE.fullmatch(part)
        if color_match is not None:
            if color is not None:
                raise ValueError(f"duplicate color filter in selector: {original_selector!r}")
            color = int(color_match.group("value"))
            continue

        raise ValueError(f"unsupported query filter in selector: {original_selector!r}")

    return layer, color


def _dxf_bbox(dxf_document: dict[str, Any]) -> dict[str, float | int] | None:
    points: list[tuple[float, float]] = []
    counted_entities = 0
    for entity in dxf_document.get("entities", []):
        entity_points = _dxf_entity_points(entity)
        if not entity_points:
            continue
        counted_entities += 1
        points.extend(entity_points)

    if not points:
        return None

    xs = [p[0] for p in points]
    ys = [p[1] for p in points]
    min_x = min(xs)
    min_y = min(ys)
    max_x = max(xs)
    max_y = max(ys)
    return {
        "min_x": min_x,
        "min_y": min_y,
        "max_x": max_x,
        "max_y": max_y,
        "width": max_x - min_x,
        "height": max_y - min_y,
        "entity_count": counted_entities,
    }


def _dxf_stats(dxf_document: dict[str, Any]) -> dict[str, Any]:
    entities = list(dxf_document.get("entities", []))
    by_type: Counter[str] = Counter()
    by_layer: Counter[str] = Counter()
    by_color: Counter[int] = Counter()

    for entity in entities:
        entity_type = str(entity.get("type", "")).upper() or "UNKNOWN"
        by_type[entity_type] += 1

        layer = str(entity.get("layer", "0")) or "0"
        by_layer[layer] += 1

        try:
            color = int(entity.get("color", 256))
        except (TypeError, ValueError):
            color = 256
        by_color[color] += 1

    return {
        "entity_count": len(entities),
        "type_count": len(by_type),
        "layer_count": len(by_layer),
        "color_count": len(by_color),
        "by_type": dict(sorted(by_type.items())),
        "by_layer": dict(sorted(by_layer.items())),
        "by_color": dict(sorted(by_color.items())),
    }


def _dxf_entity_points(entity: dict[str, Any]) -> list[tuple[float, float]]:
    try:
        entity_type = str(entity.get("type", "")).upper()
        if entity_type == "LINE":
            return [
                (_as_float(entity["x1"]), _as_float(entity["y1"])),
                (_as_float(entity["x2"]), _as_float(entity["y2"])),
            ]
        if entity_type == "POINT":
            return [(_as_float(entity["x"]), _as_float(entity["y"]))]
        if entity_type == "TEXT":
            return [(_as_float(entity["x"]), _as_float(entity["y"]))]
        if entity_type == "INSERT":
            return [(_as_float(entity["x"]), _as_float(entity["y"]))]
        if entity_type == "SOLID":
            return [
                (_as_float(entity["x1"]), _as_float(entity["y1"])),
                (_as_float(entity["x2"]), _as_float(entity["y2"])),
                (_as_float(entity["x3"]), _as_float(entity["y3"])),
                (_as_float(entity["x4"]), _as_float(entity["y4"])),
            ]
        if entity_type == "CIRCLE":
            cx = _as_float(entity["center_x"])
            cy = _as_float(entity["center_y"])
            radius = abs(_as_float(entity["radius"]))
            return [
                (cx - radius, cy),
                (cx + radius, cy),
                (cx, cy - radius),
                (cx, cy + radius),
            ]
        if entity_type == "ARC":
            return _arc_bbox_points(
                _as_float(entity["center_x"]),
                _as_float(entity["center_y"]),
                abs(_as_float(entity["radius"])),
                _as_float(entity["start_angle"]),
                _as_float(entity["end_angle"]),
            )
        if entity_type == "ELLIPSE":
            return _ellipse_bbox_points(
                _as_float(entity["center_x"]),
                _as_float(entity["center_y"]),
                _as_float(entity["major_axis_x"]),
                _as_float(entity["major_axis_y"]),
                _as_float(entity["minor_ratio"]),
                _as_float(entity["start_param"]),
                _as_float(entity["end_param"]),
            )
    except (KeyError, TypeError, ValueError):
        return []
    return []


def _as_float(value: Any) -> float:
    return float(value)


def _arc_bbox_points(
    center_x: float,
    center_y: float,
    radius: float,
    start_angle_deg: float,
    end_angle_deg: float,
) -> list[tuple[float, float]]:
    if radius <= 0.0:
        return [(center_x, center_y)]

    start = start_angle_deg % 360.0
    end = end_angle_deg % 360.0
    if end < start:
        end += 360.0

    points = [_arc_point(center_x, center_y, radius, start), _arc_point(center_x, center_y, radius, end)]
    for angle in (0.0, 90.0, 180.0, 270.0):
        candidate = angle
        while candidate < start:
            candidate += 360.0
        if start <= candidate <= end:
            points.append(_arc_point(center_x, center_y, radius, candidate))
    return points


def _arc_point(center_x: float, center_y: float, radius: float, angle_deg: float) -> tuple[float, float]:
    rad = math.radians(angle_deg)
    return (center_x + radius * math.cos(rad), center_y + radius * math.sin(rad))


def _ellipse_bbox_points(
    center_x: float,
    center_y: float,
    major_axis_x: float,
    major_axis_y: float,
    minor_ratio: float,
    start_param: float,
    end_param: float,
) -> list[tuple[float, float]]:
    start = start_param
    end = end_param
    if end <= start:
        end += 2.0 * math.pi
    span = max(0.0, end - start)

    sample_count = max(48, int(math.ceil(128.0 * (span / (2.0 * math.pi)))))
    minor_x = -major_axis_y * minor_ratio
    minor_y = major_axis_x * minor_ratio

    points = []
    for i in range(sample_count + 1):
        t = start + span * (i / sample_count)
        cos_t = math.cos(t)
        sin_t = math.sin(t)
        x = center_x + major_axis_x * cos_t + minor_x * sin_t
        y = center_y + major_axis_y * cos_t + minor_y * sin_t
        points.append((x, y))
    return points


def _normalize_max_block_nesting(max_block_nesting: int) -> int:
    value = int(max_block_nesting)
    if value < 1:
        raise ValueError("max_block_nesting must be >= 1")
    return value


def _emit_report(report: dict[str, Any], report_format: str | None, report_path: str | None) -> None:
    if report_format != "json":
        return
    if report_path:
        body = json.dumps(report, ensure_ascii=False, indent=2)
        out = Path(report_path)
        out.parent.mkdir(parents=True, exist_ok=True)
        out.write_text(body + "\n", encoding="utf-8")
    else:
        _print_json(report)


def _print_json(value: Any) -> None:
    body = json.dumps(value, ensure_ascii=False, indent=2)
    try:
        sys.stdout.write(body + "\n")
    except UnicodeEncodeError:
        fallback = json.dumps(value, ensure_ascii=True, indent=2)
        sys.stdout.write(fallback + "\n")


def _run(argv: list[str] | None = None) -> int:
    parser = _build_parser()
    args = parser.parse_args(argv)

    if args.command == "audit":
        try:
            max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        try:
            result = audit(
                args.path,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
        except Exception as exc:
            print(f"failed audit: {args.path}: {exc}", file=sys.stderr)
            return 2

        if args.json:
            _print_json(result)
        else:
            print(f"file: {args.path}")
            print(f"has_issues: {result['has_issues']}")
            print(f"issue_codes: {result['issue_codes']}")
            print(f"unresolved_count: {result['unresolved_count']}")
            print(f"unsupported_count: {result['unsupported_count']}")

        if args.fail_on_issues and result["has_issues"]:
            return 3
        return 0

    if args.command == "bbox":
        try:
            max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        try:
            result = bbox(
                args.path,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
        except Exception as exc:
            print(f"failed bbox: {args.path}: {exc}", file=sys.stderr)
            return 2

        if args.json:
            _print_json(result)
        else:
            print(f"file: {args.path}")
            if result is None:
                print("bbox: none")
            else:
                print(
                    "bbox: "
                    f"min=({result['min_x']:.6f}, {result['min_y']:.6f}) "
                    f"max=({result['max_x']:.6f}, {result['max_y']:.6f}) "
                    f"size=({result['width']:.6f}, {result['height']:.6f}) "
                    f"entities={result['entity_count']}"
                )
        return 0

    if args.command == "stats":
        try:
            max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        try:
            result = stats(
                args.path,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
        except Exception as exc:
            print(f"failed stats: {args.path}: {exc}", file=sys.stderr)
            return 2

        if args.json:
            _print_json(result)
        else:
            print(f"file: {args.path}")
            print(f"entity_count: {result['entity_count']}")
            print(f"type_count: {result['type_count']}")
            print(f"layer_count: {result['layer_count']}")
            print(f"color_count: {result['color_count']}")
            print(f"by_type: {result['by_type']}")
        return 0

    if args.command == "report":
        try:
            max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        try:
            result = report(
                args.path,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
        except Exception as exc:
            print(f"failed report: {args.path}: {exc}", file=sys.stderr)
            return 2

        if args.json:
            _print_json(result)
        else:
            audit_result = result["audit"]
            stats_result = result["stats"]
            bbox_result = result["bbox"]
            print(f"file: {args.path}")
            print(f"has_issues: {audit_result['has_issues']}")
            print(f"entity_count: {stats_result['entity_count']}")
            if bbox_result is None:
                print("bbox: none")
            else:
                print(
                    "bbox: "
                    f"min=({bbox_result['min_x']:.6f}, {bbox_result['min_y']:.6f}) "
                    f"max=({bbox_result['max_x']:.6f}, {bbox_result['max_y']:.6f})"
                )

        if args.fail_on_issues and result["audit"]["has_issues"]:
            return 3
        return 0

    if args.command == "info":
        doc = read_document(args.path)
        if args.json:
            _print_json(doc)
            return 0

        entity_total = sum(doc["entity_counts"].values())
        print(f"file: {args.path}")
        print(f"version: {doc['header']['version']}")
        print(f"entities: {entity_total}")
        print(f"block_defs: {len(doc['block_defs'])}")
        print(f"unresolved_block_refs: {doc['validation']['unresolved_def_numbers']}")
        print(f"unsupported_for_dxf: {len(read_dxf_document(args.path)['unsupported_entities'])}")
        return 0

    if args.command == "to-dxf":
        input_path = Path(args.path)
        try:
            max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        output = Path(args.output) if args.output else _default_output_path(input_path)
        output.parent.mkdir(parents=True, exist_ok=True)
        drawing: Drawing | None = None
        error: str | None = None
        try:
            drawing = readfile(str(input_path))
            drawing.saveas(
                output,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
            print(f"wrote: {output}")
            exit_code = 0
        except Exception as exc:
            error = str(exc)
            print(f"failed: {input_path} -> {output}: {error}", file=sys.stderr)
            exit_code = 2

        report_payload = {
            "source": str(input_path),
            "output": str(output),
            "ok": error is None,
            "error": error,
            "explode_inserts": bool(args.explode_inserts),
            "max_block_nesting": int(max_block_nesting),
            "audit": (
                drawing.audit(
                    explode_inserts=args.explode_inserts,
                    max_block_nesting=max_block_nesting,
                )
                if drawing is not None
                else None
            ),
        }
        _emit_report(report_payload, args.report, args.report_path)
        return exit_code

    if args.command == "plot":
        input_path = Path(args.path)
        try:
            max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        save_path: Path | None = None
        if args.output:
            save_path = Path(args.output)
        elif not args.show:
            save_path = input_path.with_suffix(".png")

        if save_path is not None:
            save_path.parent.mkdir(parents=True, exist_ok=True)

        layers = None
        if args.layers:
            layers = [layer.strip() for layer in args.layers.split(",") if layer.strip()]

        try:
            plot_jww(
                str(input_path),
                layers=layers,
                show=args.show,
                save_path=str(save_path) if save_path is not None else None,
                dpi=args.dpi,
                draw_text=not args.no_text,
                draw_points=not args.no_points,
                draw_inserts=not args.no_inserts,
                invert_y=args.invert_y,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
        except ImportError as exc:
            print(str(exc), file=sys.stderr)
            return 2

        if save_path is not None:
            print(f"wrote: {save_path}")
        elif args.show:
            print("plotted")
        return 0

    input_dir = Path(args.input_dir)
    if not input_dir.exists() or not input_dir.is_dir():
        print(f"input_dir is not a directory: {input_dir}", file=sys.stderr)
        return 2

    files = _collect_jww_files(input_dir, args.recursive)
    if not files:
        print(f"no .jww files found in {input_dir}", file=sys.stderr)
        return 1

    output_dir = Path(args.output_dir) if args.output_dir else input_dir
    try:
        max_block_nesting = _normalize_max_block_nesting(args.max_block_nesting)
    except ValueError as exc:
        print(str(exc), file=sys.stderr)
        return 2
    success = 0
    failed = 0
    report_items: list[dict[str, Any]] = []

    for src in files:
        if args.output_dir:
            rel = src.relative_to(input_dir)
            dst = (output_dir / rel).with_suffix(".dxf")
        else:
            dst = _default_output_path(src)
        dst.parent.mkdir(parents=True, exist_ok=True)

        try:
            drawing = readfile(str(src))
            drawing.saveas(
                dst,
                explode_inserts=args.explode_inserts,
                max_block_nesting=max_block_nesting,
            )
            success += 1
            report_items.append(
                {
                    "source": str(src),
                    "output": str(dst),
                    "ok": True,
                    "error": None,
                    "explode_inserts": bool(args.explode_inserts),
                    "max_block_nesting": int(max_block_nesting),
                    "audit": drawing.audit(
                        explode_inserts=args.explode_inserts,
                        max_block_nesting=max_block_nesting,
                    ),
                }
            )
        except Exception as exc:
            failed += 1
            err_text = str(exc)
            print(f"failed: {src} -> {dst}: {err_text}", file=sys.stderr)
            report_items.append(
                {
                    "source": str(src),
                    "output": str(dst),
                    "ok": False,
                    "error": err_text,
                    "explode_inserts": bool(args.explode_inserts),
                    "max_block_nesting": int(max_block_nesting),
                    "audit": None,
                }
            )
            if args.fail_fast:
                report_payload = {
                    "input_dir": str(input_dir),
                    "output_dir": str(output_dir),
                    "recursive": bool(args.recursive),
                    "explode_inserts": bool(args.explode_inserts),
                    "max_block_nesting": int(max_block_nesting),
                    "converted": success,
                    "failed": failed,
                    "items": report_items,
                }
                _emit_report(report_payload, args.report, args.report_path)
                return 2

    print(f"converted={success} failed={failed} output_dir={output_dir}")
    report_payload = {
        "input_dir": str(input_dir),
        "output_dir": str(output_dir),
        "recursive": bool(args.recursive),
        "explode_inserts": bool(args.explode_inserts),
        "max_block_nesting": int(max_block_nesting),
        "converted": success,
        "failed": failed,
        "items": report_items,
    }
    _emit_report(report_payload, args.report, args.report_path)
    return 0 if failed == 0 else 2


def main() -> None:
    raise SystemExit(_run())
